==> ask-cli-1.5.0.tgz: package/lib/deploy/skill-schema-pocessor.js <==
const jsonfile = require('jsonfile');
const async = require('async');

const CONSTANT = require('@src/utils/constants');
const urlUtils = require('@src/utils/url-utils');
const lambdaEndpointUtility = require('@src/utils/lambda-operation/lambda-endpoint-utility');
const parser = require('../utils/skill-parser');
const profileHelper = require('../utils/profile-helper');
const jsonUtility = require('../utils/json-utility');
const domainRegistry = require('../utils/domain-registry');
const retrieveLambda = require('../utils/lambda-operation/retrieve-lambda');
const tools = require('../utils/tools');

// Public
module.exports = {
    updateSkillSchemaWithLambdaCreation,
    parseSkill
};

function updateSkillSchemaWithLambdaCreation(generatedLambdaList, submittingReadySkillSchema) {
    const finalSkillSchema = JSON.parse(JSON.stringify(submittingReadySkillSchema));
    if (!generatedLambdaList || generatedLambdaList.length === 0) {
        return finalSkillSchema;
    }
    for (const metaData of generatedLambdaList) {
        for (const usage of metaData.alexaUsage) {
            const domain = domainRegistry.getSkillSchemaKey(usage.split('/')[0]);
            const region = usage.split('/')[1];
            const arrayPathToUri = ['manifest', 'apis', domain, 'regions', region, 'endpoint', 'uri'];
            jsonUtility.insertObjectToObject(finalSkillSchema, arrayPathToUri, metaData.arn);
        }
    }
    // Logic to set one region as default.
    for (const domain of Object.keys(finalSkillSchema.manifest.apis)) {
        const domainInfo = finalSkillSchema.manifest.apis[domain];
        if (domainInfo.endpoint) {
            // If default region existed and need to create Lambda, replace it with real ARN
            if (domainInfo.regions.default) {
                domainInfo.endpoint.uri = domainInfo.regions.default.endpoint.uri;
                jsonUtility.deletePropertyFromJsonObject(domainInfo, ['regions', 'default']);
            }
        } else {
            // If NA exists move NA up; else move whatever comes the first
            const regionsKeyArray = Object.keys(domainInfo.regions);
            if (regionsKeyArray.indexOf('NA') === -1) {
                const firstRegion = domainInfo.regions[regionsKeyArray[0]];
                const defaultEndpoint = firstRegion.endpoint;
                domainInfo.endpoint = defaultEndpoint;
                jsonUtility.deletePropertyFromJsonObject(domainInfo, ['regions', regionsKeyArray[0]]);
            } else {
                const defaultEndpoint = domainInfo.regions.NA;
                domainInfo.endpoint = defaultEndpoint;
                jsonUtility.deletePropertyFromJsonObject(domainInfo, ['regions', 'NA']);
            }
        }
        // Clean empty regions if no endpoint inside
        if (Object.keys(domainInfo.regions) === 0) {
            jsonUtility.deletePropertyFromJsonObject(domainInfo, ['regions']);
        }
    }
    return finalSkillSchema;
}

function parseSkill(projectConfigFile, skillFile, profile, callback) {
    const manifest = jsonUtility.read(skillFile);
    if (!isValidManifest(manifest)) {
        console.error('[Error]: Invalid skill manifest. Make sure your skill.json always contains "manifest" field.');
        return;
    }

    const askConfig = jsonUtility.read(projectConfigFile);
    if (!askConfig.deploy_settings[profile] || Object.keys(askConfig.deploy_settings[profile]).length === 0) {
        addNewProfileToConfig(askConfig, profile);
        jsonfile.writeFileSync(projectConfigFile, askConfig, { spaces: 2 });
    }

    const skillId = jsonUtility.getPropertyValueFromObject(askConfig, ['deploy_settings', profile, 'skill_id']);

    // doing merge
    const mergedManifest = tools.collate(manifest, askConfig.deploy_settings[profile].merge);
    // move default endpoint under regions to treat it as regular endpoint
    moveDefaultEndpointIntoRegions(mergedManifest);
    // create file
    generateSubmittingReadyManifest(mergedManifest, profile, (preprocessedManifest) => {
        // this Manifest has sourceDir and url -> either true url, or functionName that need to be created.
        // which means if there is a functionName that can be find on AWS, it will change to the ARN.
        const withTrueURLManifest = tools.collate(mergedManifest, preprocessedManifest);

        moveDefaultEndpointUpBack(withTrueURLManifest);
        moveDefaultEndpointUpBack(preprocessedManifest);

        const skillInfo = parser.extractSkillInfo(withTrueURLManifest);

        callback(preprocessedManifest, skillInfo, skillId);
    });
}

function moveDefaultEndpointIntoRegions(mergedManifest) {
    const apisInfo = jsonUtility.getPropertyValueFromObject(mergedManifest, ['manifest', 'apis']);
    for (const domain of Object.getOwnPropertyNames(apisInfo)) {
        const domainInfo = apisInfo[domain];
        if (domainInfo.endpoint) {
            if (!domainInfo.regions) {
                domainInfo.regions = {};
            }
            domainInfo.regions.default = {
                endpoint: domainInfo.endpoint
            };
        }
    }
    return mergedManifest;
}

function generateSubmittingReadyManifest(mergedManifest, profile, callback) {
    const deepCopiedManifest = JSON.parse(JSON.stringify(mergedManifest));
    const awsProfile = profileHelper.getAWSProfile(profile);
    async.each(domainRegistry.domainList(), (domain, domainAsyncCallback) => {
        const domainKey = domainRegistry.getSkillSchemaKey(domain);
        const pathToRegions = ['manifest', 'apis', domainKey, 'regions'];
        const regionsInfo = jsonUtility.getPropertyValueFromObject(deepCopiedManifest, pathToRegions);
        if (!regionsInfo || Object.getOwnPropertyNames(regionsInfo).length === 0) {
            domainAsyncCallback();
            return;
        }
        async.each(Object.getOwnPropertyNames(regionsInfo), (region, regionLevelCallback) => {
            const currentUri = regionsInfo[region].endpoint.uri;
            const currentSourceDir = regionsInfo[region].endpoint.sourceDir;

            // 1.Validate current uri (after merge) exists
            if (typeof currentUri !== 'string' || currentUri.trim().length === 0) {
                regionLevelCallback(`[Error]: Please provide the uri (either ARN or function name if using Lambda) for ${domain}/${region}.`);
                return;
            }
            // 2.uri is https endpoint
            if (urlUtils.isHttpsUrl(currentUri)) {
                if (typeof currentSourceDir === 'string' && currentSourceDir.trim().length !== 0) {
                    regionLevelCallback(`[Error]: CLI doesn't support sourceDir deploy for HTTPS endpoint for ${domain}/${region}.`);
                } else {
                    regionLevelCallback();
                }
                return;
            }
            // 3.uri is Lambda ARN or function name
            if (!awsProfile) {
                if (typeof currentSourceDir === 'string' && currentSourceDir.trim().length !== 0) {
                    // when user doesn't set awsProfile but still provides sourceDir, delete the region and warn out
                    console.warn(`[Warn]: AWS profile not found, thus delete the Lambda endpoint for ${domain}/${region}`);
                    delete regionsInfo[region];
                }
                regionLevelCallback();
                return;
            }
            if (urlUtils.isLambdaArn(currentUri)) {
                // 3.1 current uri is Lambda ARN
                const { awsRegion } = lambdaEndpointUtility.parseLambdaArn(currentUri);
                retrieveLambda.retrieveLambdaFunction(currentUri, awsRegion, awsProfile, (err, res) => {
                    if (err) {
                        regionLevelCallback(`[Error]: Get Lambda ARN ${currentUri} failure for ${domain}/${region}.\n${err}`);
                        return;
                    }
                    if (!res) {
                        regionLevelCallback(`[Error]: The Lambda ARN ${currentUri} doesn't exist for ${domain}/${region}.`);
                        return;
                    }
                    regionLevelCallback();
                });
            } else {
                // 3.2 current uri is Lambda function name
                const awsRegion = CONSTANT.REGION_MAPPING[region] || CONSTANT.REGION_MAPPING.NA;
                retrieveLambda.retrieveLambdaFunction(currentUri, awsRegion, awsProfile, (err, res) => {
                    if (err) {
                        regionLevelCallback(`[Error]: Get Lambda ARN ${currentUri} failure for ${domain}/${region}.\n${err}`);
                        return;
                    }
                    if (!res) {
                        if (typeof currentSourceDir === 'string' && currentSourceDir.trim().length !== 0) {
                            // this is a valid case if Lambda function does not exist
                            regionLevelCallback();
                        } else {
                            regionLevelCallback(`[Error]: Failed to resolve Lambda endpoint for ${domain}/${region}.
Please specify sourceDir if you want to create Lambda with the function name, or provide valid Lambda ARN if you only want to set the endpoint.`);
                        }
                    } else {
                        // if Lambda function exists, replace it with true Lambda ARN
                        const awsResponseURL = res.Configuration.FunctionArn;
                        if (urlUtils.isLambdaArn(awsResponseURL)) {
                            const arrayPathToUri = [region, 'endpoint', 'uri'];
                            jsonUtility.insertObjectToObject(regionsInfo, arrayPathToUri, awsResponseURL);
                            regionLevelCallback();
                        } else {
                            regionLevelCallback(`[Error]: Please make sure you use correct Lambda function name for ${domain}/${region}.`);
                        }
                    }
                });
            }
        }, (err) => {
            if (err) {
                domainAsyncCallback(err);
            }
            domainAsyncCallback();
        });
    }, (err) => {
        if (err) {
            console.error(err);
            process.exit(1);
        }
        obtainUniqueness(deepCopiedManifest, (approvedManifest) => {
            callback(approvedManifest);
        });
    });
}

// check whether multiple different code bases map to a single lambda uri, and remove 'sourceDir' and pass to callback.
function obtainUniqueness(manifest, callback) {
    const checkUniqueMap = new Map();
    const domainKeyArray = domainRegistry.domainList().map(domain => domainRegistry.getSkillSchemaKey(domain));

    for (const domainKey of domainKeyArray) {
        if (!manifest.manifest.apis[domainKey] || !manifest.manifest.apis[domainKey].regions) {
            continue;
        }
        const regionsInfo = manifest.manifest.apis[domainKey].regions;
        for (const region of Object.getOwnPropertyNames(regionsInfo)) {
            const { uri } = regionsInfo[region].endpoint;
            if (!uri) {
                continue;
            }
            const deletePath = ['manifest', 'apis', domainKey, 'regions', region, 'endpoint', 'sourceDir'];
            if (uri.substr(0, 4) !== 'arn:') {
                // it's a function name, which aws don't have a function match with
                jsonUtility.deletePropertyFromJsonObject(manifest, deletePath);
                continue;
            }
            const { sourceDir } = regionsInfo[region].endpoint;
            if (!sourceDir || sourceDir.length === 0) {
                continue;
            }
            if (!checkUniqueMap.get(uri)) {
                checkUniqueMap.set(uri, sourceDir);
            } else if (sourceDir !== checkUniqueMap.get(uri)) {
                console.error(`[Error]: Multiple sourceDir pointing to the same endpoint ${uri}.`);
                process.exit(1);
            }
            jsonUtility.deletePropertyFromJsonObject(manifest, deletePath);
        }
    }
    callback(manifest);
}

function moveDefaultEndpointUpBack(manifest) {
    const apisInfo = jsonUtility.getPropertyValueFromObject(manifest, ['manifest', 'apis']);
    for (const domain of Object.getOwnPropertyNames(apisInfo)) {
        if (apisInfo[domain].regions) {
            apisInfo[domain].endpoint = jsonUtility.getPropertyValueFromObject(
                apisInfo, [domain, 'regions', 'default', 'endpoint']
            );
            jsonUtility.deletePropertyFromJsonObject(apisInfo, [domain, 'regions', 'default']);
            // Remove "regions" field if there is no region inside
            if (Object.keys(apisInfo[domain].regions).length === 0) {
                jsonUtility.deletePropertyFromJsonObject(apisInfo, [domain, 'regions']);
            }
        }
    }
}

function addNewProfileToConfig(askConfig, profile) {
    askConfig.deploy_settings[profile] = {
        skill_id: '',
        was_cloned: false,
        merge: {}
    };
}

// a basic validation for skill.json
// to check whether it has manifest or not
function isValidManifest(skillSchema) {
    return skillSchema && skillSchema.manifest && typeof skillSchema.manifest === 'object';
}

==> ask-cli-1.6.0.tgz: package/lib/deploy/skill-schema-pocessor.js <==
const jsonfile = require('jsonfile');
const async = require('async');

const CONSTANT = require('@src/utils/constants');
const urlUtils = require('@src/utils/url-utils');
const lambdaEndpointUtility = require('@src/utils/lambda-operation/lambda-endpoint-utility');
const parser = require('../utils/skill-parser');
const profileHelper = require('../utils/profile-helper');
const jsonUtility = require('../utils/json-utility');
const domainRegistry = require('../utils/domain-registry');
const retrieveLambda = require('../utils/lambda-operation/retrieve-lambda');
const tools = require('../utils/tools');

// Public
module.exports = {
    updateSkillSchemaWithLambdaCreation,
    parseSkill
};

function updateSkillSchemaWithLambdaCreation(generatedLambdaList, submittingReadySkillSchema) {
    const finalSkillSchema = JSON.parse(JSON.stringify(submittingReadySkillSchema));
    if (!generatedLambdaList || generatedLambdaList.length === 0) {
        return finalSkillSchema;
    }
    for (const metaData of generatedLambdaList) {
        for (const usage of metaData.alexaUsage) {
            const domain = domainRegistry.getSkillSchemaKey(usage.split('/')[0]);
            const region = usage.split('/')[1];
            const arrayPathToUri = ['manifest', 'apis', domain, 'regions', region, 'endpoint', 'uri'];
            jsonUtility.insertObjectToObject(finalSkillSchema, arrayPathToUri, metaData.arn);
        }
    }
    // Logic to set one region as default.
    for (const domain of Object.keys(finalSkillSchema.manifest.apis)) {
        const domainInfo = finalSkillSchema.manifest.apis[domain];
        if (domainInfo.endpoint) {
            // If default region existed and need to create Lambda, replace it with real ARN
            if (domainInfo.regions.default) {
                domainInfo.endpoint.uri = domainInfo.regions.default.endpoint.uri;
                jsonUtility.deletePropertyFromJsonObject(domainInfo, ['regions', 'default']);
            }
        } else {
            // If NA exists move NA up; else move whatever comes the first
            const regionsKeyArray = Object.keys(domainInfo.regions);
            if (regionsKeyArray.indexOf('NA') === -1) {
                const firstRegion = domainInfo.regions[regionsKeyArray[0]];
                const defaultEndpoint = firstRegion.endpoint;
                domainInfo.endpoint = defaultEndpoint;
                jsonUtility.deletePropertyFromJsonObject(domainInfo, ['regions', regionsKeyArray[0]]);
            } else {
                const defaultEndpoint = domainInfo.regions.NA;
                domainInfo.endpoint = defaultEndpoint;
                jsonUtility.deletePropertyFromJsonObject(domainInfo, ['regions', 'NA']);
            }
        }
        // Clean empty regions if no endpoint inside
        if (Object.keys(domainInfo.regions) === 0) {
            jsonUtility.deletePropertyFromJsonObject(domainInfo, ['regions']);
        }
    }
    return finalSkillSchema;
}

function parseSkill(projectConfigFile, skillFile, profile, callback) {
    const manifest = jsonUtility.read(skillFile);
    if (!isValidManifest(manifest)) {
        console.error('[Error]: Invalid skill manifest. Make sure your skill.json always contains "manifest" field.');
        return;
    }

    const askConfig = jsonUtility.read(projectConfigFile);
    if (!askConfig.deploy_settings[profile] || Object.keys(askConfig.deploy_settings[profile]).length === 0) {
        addNewProfileToConfig(askConfig, profile);
        jsonfile.writeFileSync(projectConfigFile, askConfig, { spaces: 2 });
    }

    const skillId = jsonUtility.getPropertyValueFromObject(askConfig, ['deploy_settings', profile, 'skill_id']);

    // doing merge
    const mergedManifest = tools.collate(manifest, askConfig.deploy_settings[profile].merge);
    // move default endpoint under regions to treat it as regular endpoint
    moveDefaultEndpointIntoRegions(mergedManifest);
    // create file
    generateSubmittingReadyManifest(mergedManifest, profile, (preprocessedManifest) => {
        // this Manifest has sourceDir and url -> either true url, or functionName that need to be created.
        // which means if there is a functionName that can be find on AWS, it will change to the ARN.
        const withTrueURLManifest = tools.collate(mergedManifest, preprocessedManifest);

        moveDefaultEndpointUpBack(withTrueURLManifest);
        moveDefaultEndpointUpBack(preprocessedManifest);

        const skillInfo = parser.extractSkillInfo(withTrueURLManifest);

        callback(preprocessedManifest, skillInfo, skillId);
    });
}

function moveDefaultEndpointIntoRegions(mergedManifest) {
    const apisInfo = jsonUtility.getPropertyValueFromObject(mergedManifest, ['manifest', 'apis']);
    for (const domain of Object.getOwnPropertyNames(apisInfo)) {
        const domainInfo = apisInfo[domain];
        if (domainInfo.endpoint) {
            if (!domainInfo.regions) {
                domainInfo.regions = {};
            }
            domainInfo.regions.default = {
                endpoint: domainInfo.endpoint
            };
        }
    }
    return mergedManifest;
}

function generateSubmittingReadyManifest(mergedManifest, profile, callback) {
    const deepCopiedManifest = JSON.parse(JSON.stringify(mergedManifest));
    const awsProfile = profileHelper.getAWSProfile(profile);
    async.each(domainRegistry.domainList(), (domain, domainAsyncCallback) => {
        const domainKey = domainRegistry.getSkillSchemaKey(domain);
        const pathToRegions = ['manifest', 'apis', domainKey, 'regions'];
        const regionsInfo = jsonUtility.getPropertyValueFromObject(deepCopiedManifest, pathToRegions);
        if (!regionsInfo || Object.getOwnPropertyNames(regionsInfo).length === 0) {
            domainAsyncCallback();
            return;
        }
        async.each(Object.getOwnPropertyNames(regionsInfo), (region, regionLevelCallback) => {
            const currentUri = regionsInfo[region].endpoint.uri;
            const currentSourceDir = regionsInfo[region].endpoint.sourceDir;

            // 1.Validate current uri (after merge) exists
            if (typeof currentUri !== 'string' || currentUri.trim().length === 0) {
                regionLevelCallback(`[Error]: Please provide the uri (either ARN or function name if using Lambda) for ${domain}/${region}.`);
                return;
            }
            // 2.uri is https endpoint
            if (urlUtils.isHttpsUrl(currentUri)) {
                if (typeof currentSourceDir === 'string' && currentSourceDir.trim().length !== 0) {
                    regionLevelCallback(`[Error]: CLI doesn't support sourceDir deploy for HTTPS endpoint for ${domain}/${region}.`);
                } else {
                    regionLevelCallback();
                }
                return;
            }
            // 3.uri is Lambda ARN or function name
            if (!awsProfile) {
                if (typeof currentSourceDir === 'string' && currentSourceDir.trim().length !== 0) {
                    // when user doesn't set awsProfile but still provides sourceDir, delete the region and warn out
                    console.warn(`[Warn]: AWS profile not found, thus delete the Lambda endpoint for ${domain}/${region}`);
                    delete regionsInfo[region];
                }
                regionLevelCallback();
                return;
            }
            if (urlUtils.isLambdaArn(currentUri)) {
                // 3.1 current uri is Lambda ARN
                const awsRegion = lambdaEndpointUtility.resolveLambdaAwsRegion(currentUri, region, awsProfile);
                retrieveLambda.retrieveLambdaFunction(currentUri, awsRegion, awsProfile, (err, res) => {
                    if (err) {
                        regionLevelCallback(`[Error]: Get Lambda ARN ${currentUri} failure for ${domain}/${region}.\n${err}`);
                        return;
                    }
                    if (!res) {
                        regionLevelCallback(`[Error]: The Lambda ARN ${currentUri} doesn't exist for ${domain}/${region}.`);
                        return;
                    }
                    regionLevelCallback();
                });
            } else {
                // 3.2 current uri is Lambda function name
                const awsRegion = lambdaEndpointUtility.resolveLambdaAwsRegion(null, region, awsProfile);
                retrieveLambda.retrieveLambdaFunction(currentUri, awsRegion, awsProfile, (err, res) => {
                    if (err) {
                        regionLevelCallback(`[Error]: Get Lambda ARN ${currentUri} failure for ${domain}/${region}.\n${err}`);
                        return;
                    }
                    if (!res) {
                        if (typeof currentSourceDir === 'string' && currentSourceDir.trim().length !== 0) {
                            // this is a valid case if Lambda function does not exist
                            regionLevelCallback();
                        } else {
                            regionLevelCallback(`[Error]: Failed to resolve Lambda endpoint for ${domain}/${region}.
Please specify sourceDir if you want to create Lambda with the function name, or provide valid Lambda ARN if you only want to set the endpoint.`);
                        }
                    } else {
                        // if Lambda function exists, replace it with true Lambda ARN
                        const awsResponseURL = res.Configuration.FunctionArn;
                        if (urlUtils.isLambdaArn(awsResponseURL)) {
                            const arrayPathToUri = [region, 'endpoint', 'uri'];
                            jsonUtility.insertObjectToObject(regionsInfo, arrayPathToUri, awsResponseURL);
                            regionLevelCallback();
                        } else {
                            regionLevelCallback(`[Error]: Please make sure you use correct Lambda function name for ${domain}/${region}.`);
                        }
                    }
                });
            }
        }, (err) => {
            if (err) {
                domainAsyncCallback(err);
            }
            domainAsyncCallback();
        });
    }, (err) => {
        if (err) {
            console.error(err);
            process.exit(1);
        }
        obtainUniqueness(deepCopiedManifest, (approvedManifest) => {
            callback(approvedManifest);
        });
    });
}

// check whether multiple different code bases map to a single lambda uri, and remove 'sourceDir' and pass to callback.
function obtainUniqueness(manifest, callback) {
    const checkUniqueMap = new Map();
    const domainKeyArray = domainRegistry.domainList().map(domain => domainRegistry.getSkillSchemaKey(domain));

    for (const domainKey of domainKeyArray) {
        if (!manifest.manifest.apis[domainKey] || !manifest.manifest.apis[domainKey].regions) {
            continue;
        }
        const regionsInfo = manifest.manifest.apis[domainKey].regions;
        for (const region of Object.getOwnPropertyNames(regionsInfo)) {
            const { uri } = regionsInfo[region].endpoint;
            if (!uri) {
                continue;
            }
            const deletePath = ['manifest', 'apis', domainKey, 'regions', region, 'endpoint', 'sourceDir'];
            if (uri.substr(0, 4) !== 'arn:') {
                // it's a function name, which aws don't have a function match with
                jsonUtility.deletePropertyFromJsonObject(manifest, deletePath);
                continue;
            }
            const { sourceDir } = regionsInfo[region].endpoint;
            if (!sourceDir || sourceDir.length === 0) {
                continue;
            }
            if (!checkUniqueMap.get(uri)) {
                checkUniqueMap.set(uri, sourceDir);
            } else if (sourceDir !== checkUniqueMap.get(uri)) {
                console.error(`[Error]: Multiple sourceDir pointing to the same endpoint ${uri}.`);
                process.exit(1);
            }
            jsonUtility.deletePropertyFromJsonObject(manifest, deletePath);
        }
    }
    callback(manifest);
}

function moveDefaultEndpointUpBack(manifest) {
    const apisInfo = jsonUtility.getPropertyValueFromObject(manifest, ['manifest', 'apis']);
    for (const domain of Object.getOwnPropertyNames(apisInfo)) {
        if (apisInfo[domain].regions) {
            apisInfo[domain].endpoint = jsonUtility.getPropertyValueFromObject(
                apisInfo, [domain, 'regions', 'default', 'endpoint']
            );
            jsonUtility.deletePropertyFromJsonObject(apisInfo, [domain, 'regions', 'default']);
            // Remove "regions" field if there is no region inside
            if (Object.keys(apisInfo[domain].regions).length === 0) {
                jsonUtility.deletePropertyFromJsonObject(apisInfo, [domain, 'regions']);
            }
        }
    }
}

function addNewProfileToConfig(askConfig, profile) {
    askConfig.deploy_settings[profile] = {
        skill_id: '',
        was_cloned: false,
        merge: {}
    };
}

// a basic validation for skill.json
// to check whether it has manifest or not
function isValidManifest(skillSchema) {
    return skillSchema && skillSchema.manifest && typeof skillSchema.manifest === 'object';
}
